/*****************************************************************************************
 ** File     : LDAP.cpp
 ** Date     : 01/31/2002
 ** Author   : Osama Dengler <osama@denglernet.de>
 ** Abstract : LDAP functions for the Unix Password sync Package
 **
 ** Copyright (C) Osama Dengler 2002
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation; either version 2 of the License, or
 ** (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program; if not, write to the Free Software
 ** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************************/
#include "stdafx.h"
#include "PasswordFilter.h"

extern PTSTR pLdapServer;
extern PTSTR pBindDN;
extern PTSTR pBindPW;
extern PTSTR pBaseDN;
extern BOOL  bUseSSL;
extern BOOL  bForceLDAPv2;
extern BOOL  bForceLDAPv3;
extern int   iLdapVersion;

/*************************************************************************************************
 ** Function  : LdapChangePassword
 ** Purpose   : Change the user's password in the LDAP directory
 ** Author    : O. Dengler
 ** History   : 03/09/2002 : 1st
 ************************************************************************************************/
BOOL LdapChangePassword( PWSTR pUserName, PWSTR pUserPassword)
{
	LDAP *ld;						// LDAP connection handle
	BOOL RetVal=TRUE;
	PWSTR pDN=NULL;

	DEBUGMSG("+LdapChangePassword\n");

	/*
	 * establish a connection to the server
	 */
	if((ld=LdapConnect()) == NULL)
	{
		DEBUGMSG("Unable to connect to LDAP server.\n");
		RetVal=FALSE;
		goto ErrorReturn;
	}

	/*
	 * bind to the directory as administrator
	 */
	if(!LdapBind(ld))
	{
		DEBUGMSG("Unable to bind to the LDAP directory.\n");
		RetVal=FALSE;
		goto ErrorReturn;
	}

	/*
	 * find the user's account in the directory
	 */
	if((pDN=LdapFindUserAccount(ld, pUserName)) == NULL)
	{
		DEBUGMSG("Unable to find user account in the LDAP directory.\n");
		RetVal=FALSE;
		goto ErrorReturn;
	}

	/*
	 * set new user password
	 */
	if(!LdapSetUserPassword(ld, pDN, pUserPassword))
	{
		DEBUGMSG("Unable to set user password.\n");
		RetVal=FALSE;
		goto ErrorReturn;
	}

ErrorReturn:

	/*
	 * free allocated memory
	 */
	if(!LdapFreeMemory(pDN))
	{
		DEBUGMSG("Unable to free allocated memory.\n");
		RetVal=FALSE;
	}

	/*
	 * close connection to the server
	 */
	if(!LdapClose(ld))
	{
		DEBUGMSG("Unable to close connection to the server.\n");
		RetVal=FALSE;
	}

	return(RetVal);
}

/*************************************************************************************************
 ** Function  : LdapSetPassword
 ** Purpose   : Change the user's password in the LDAP directory
 ** Author    : O. Dengler
 ** History   : 03/09/2002 : 1st
 ************************************************************************************************/
BOOL LdapSetPassword( PWSTR pUserName, PWSTR pUserPassword)
{
	BOOL Status;

	DEBUGMSG("   +LdapSetPassword\n");
	Status=LdapChangePassword(pUserName, pUserPassword);
	DEBUGMSG("   +LdapSetPassword\n");

	return(Status);
}

/*************************************************************************************************
 ** Function  : LdapConnect
 ** Purpose   : Connect to the LDAP server
 ** Author    : O. Dengler
 ** History   : 03/09/2002 : 1st
 ************************************************************************************************/
LDAP* LdapConnect()
{
	LDAP* ld;
	ULONG Status;

	DEBUGMSG("   +LdapConnect\n");

	if(LDAP_VERSION2 == iLdapVersion)
	{
		DEBUGMSG("      Using LDAPv2\n");

		ld=ldap_open(pLdapServer, LDAP_PORT);
		DEBUGMSG("      ldap_open(%S) returned \"%S\"\n",pLdapServer, ldap_err2string(LdapGetLastError()));
		if(ld==NULL)
		{
			DEBUGMSG("      Unable to connect to LDAP server. ld= 0x%08x\n",ld);
			DEBUGMSG("   -LdapConnect\n");
			return(NULL);
		}
	}
	else
	{
		DEBUGMSG("      Using LDAPv3\n");

		/*
		 * initialize the connection structure but don't yet connect.
		 */
		ld=ldap_init(pLdapServer, LDAP_PORT);
		DEBUGMSG("      ldap_init(%S) returned \"%S\"\n",pLdapServer, ldap_err2string(LdapGetLastError()));
		if(NULL == ld)
		{
			DEBUGMSG("      Unable to initialize LDAP connection. ld= 0x%08x\n",ld);
			DEBUGMSG("   -LdapConnect\n");
			return(NULL);
		}

		/*
		 * Select LDAP protocol version 3
		 */
		Status=ldap_set_option(ld, LDAP_OPT_PROTOCOL_VERSION, &iLdapVersion );
		DEBUGMSG("      ldap_set_option returned \"%S\"\n", ldap_err2string(LdapGetLastError()));
		if(LDAP_SUCCESS != Status)
		{
			DEBUGMSG( "      Unable to set LDAP_OPT_PROTOCOL_VERSION=%u\n", iLdapVersion );
			return(NULL);
		}

	}

	DEBUGMSG("      Connected to LDAP server %S\n", pLdapServer);
	DEBUGMSG("   -LdapConnect\n");
	return(ld);
}

/*************************************************************************************************
 ** Function  : LdapBind
 ** Purpose   : Bind to the LDAP directory
 ** Author    : O. Dengler
 ** History   : 03/09/2002 : 1st
 ************************************************************************************************/
BOOL LdapBind(LDAP* ld)
{
	ULONG Status;

	DEBUGMSG("   +LdapBind\n");
	DEBUGMSG("      Bind DN     : %S\n",pBindDN);

	/*
	 * bind to the server
	 */
	Status=ldap_simple_bind_s(ld, pBindDN, pBindPW);
	DEBUGMSG("      ldap_simple_bind_s returned \"%S\"\n", ldap_err2string(Status));
	if(LDAP_SUCCESS != Status)
	{
		DEBUGMSG("   -LdapBind\n");
		return(FALSE);
	}

	DEBUGMSG("   -LdapBind\n");
	return(TRUE);
}

/*************************************************************************************************
 ** Function  : LdapFindUserAccount
 ** Purpose   : Find the user's account in the LDAP directory.
 ** Author    : O. Dengler
 ** History   : 03/09/2002 : 1st
 ************************************************************************************************/
PWSTR LdapFindUserAccount(LDAP* ld, PWSTR pUserName)
{
	LDAPMessage* pMsg=NULL;
	BerElement* Ber=NULL;
	ULONG Status;
	PWSTR attr=NULL;
	PWSTR* Values=NULL;
	PWSTR pSearchString=NULL;
	PWSTR pDN=NULL;
	ULONG cbSearchString=0;

	DEBUGMSG("   +LdapFindUserAccount\n");

	/*
	 * Determinte the amount of memory necessary to hold the search string.
	 */
	cbSearchString= wcslen(_T("cn=")) * sizeof(WCHAR) +	// header
					wcslen(pUserName) * sizeof(WCHAR) +	// user name
										sizeof(WCHAR);	// trailing zero
	
	/*
	 * Allocate the search string.
	 */
	pSearchString=(PWSTR)LocalAlloc(LPTR, cbSearchString);
	DEBUGMSG("      Allocated %u Bytes @ 0x%08x for pSearchString\n", cbSearchString, pSearchString);

	/*
	 * Form the search string in the previously allocated buffer.
	 */
	wsprintf(pSearchString, _T("cn=%s"), pUserName);
	DEBUGMSG("      Search String = \"%S\"\n", pSearchString);

	/*
	 * Search for the user's account.
	 */
	Status=ldap_search_s(ld, pBaseDN, LDAP_SCOPE_ONELEVEL, pSearchString, NULL, FALSE, &pMsg);
	DEBUGMSG("      ldap_search_s returned \"%S\"\n", ldap_err2string(Status));
	if(LDAP_SUCCESS != Status)
	{
		DEBUGMSG("   -LdapFindUserAccount\n");
		goto ErrorReturn;
	}

	/*
	 * Check if no entry was found.
	 */
	pMsg = ldap_first_entry(ld, pMsg);
	if(NULL == pMsg)
	{
		DEBUGMSG("      No directory entries found.\n");
		goto ErrorReturn;
	}

	/*
	 * Get the entry's DN.
	 */
	pDN=ldap_get_dn(ld, pMsg);
	if(NULL == pDN)
	{
		DEBUGMSG("      Unable to get entry's DN. Error is \"%S\"\n",ldap_err2string(LdapGetLastError()));
		goto ErrorReturn;
	}

	/*
	 * Dump the directory entry's attributes and their values.
	 */
	DEBUGMSG("      Attributes for DN=%S\n", pDN);
	DEBUGMSG("      =====================================================\n");
	for(attr = ldap_first_attribute(ld, pMsg, &Ber);
		attr;
        attr = ldap_next_attribute(ld, pMsg, Ber))
	{
		Values=ldap_get_values(ld, pMsg, attr);
		DEBUGMSG("         %S = %S\n", attr, Values[0]);
		ldap_value_free(Values);
	}
	DEBUGMSG("      =====================================================\n");

	/*
	 * Check if more than one entry was found.
	 */
	pMsg=ldap_next_entry(ld, pMsg);
	if(NULL != pMsg)
	{
		DEBUGMSG("      More than one directory entries found\n");
		goto ErrorReturn;
	}

ErrorReturn:

	/*
	 * free allocated memory
	 */
	if(NULL != pSearchString ) LocalFree(pSearchString);

	DEBUGMSG("   -LdapFindUserAccount\n");
	return(pDN);
}

/*************************************************************************************************
 ** Function  : LdapSetUserPassword
 ** Purpose   : Set the UserPassword attribute
 ** Author    : O. Dengler
 ** History   : 03/09/2002 : 1st
 ************************************************************************************************/
BOOL LdapSetUserPassword(LDAP* ld, PWSTR pDN, PWSTR pUserPassword)
{
	DEBUGMSG("   +LdapSetUserPassword\n");

	if(LDAP_VERSION2 == iLdapVersion)
	{
		DEBUGMSG("      Using LDAPv2\n");

		/*
		 * LDAPv2 doesn't support ldap_extended_operation. Therefore we modify the UserPassword
		 * attribute ourself. Currently the password is stored unencrypted!
		 */
		ULONG Status;
		LDAPMod mod;
		LDAPMod *mods[2]={ &mod, NULL};
		PWSTR Values[2]={ pUserPassword, NULL };

		mod.mod_op=LDAP_MOD_REPLACE;
		mod.mod_type=_T("UserPassword");
		mod.mod_vals.modv_strvals=Values;

		DEBUGMSG("      Setting %S to \"%S\"\n", mods[0]->mod_type, mods[0]->mod_vals.modv_strvals);

		Status=ldap_modify_s(ld, pDN, mods);
		DEBUGMSG("      ldap_modify_s returned \"%S\"\n", ldap_err2string(Status));
		if(LDAP_SUCCESS != Status)
		{
			DEBUGMSG("      Unable to set password via ldap_modify_s\n");
			return(FALSE);
		}
	}
	else
	{
		ULONG Status, id;
		LDAPMessage * res;
		BerElement *ber;
		berval *bv;
		PSTR pcDN=NULL;
		PSTR pcPwd=NULL;

		DEBUGMSG("      Using LDAPv3\n");

		/*
		 * Create the ANSI version of the UNICODE DN.
		 */
		pcDN=(PSTR)LocalAlloc(LPTR, wcslen(pDN) +1);
		sprintf(pcDN, "%S", pDN);
		DEBUGMSG("      pcDN = %s\n", pcDN);

		/*
		 * Create the ANSI version of the UNICODE user password.
		 */
		pcPwd=(PSTR)LocalAlloc(LPTR, wcslen(pUserPassword) + 1);
		sprintf(pcPwd, "%S", pUserPassword);
		DEBUGMSG("      pcPwd = %s\n", pcPwd);

		/*
		 * Allocate the BerElement structure.
		 */
		ber=ber_alloc_t(LBER_USE_DER);
		if(NULL == ber)
		{
			DEBUGMSG("      Unable to allocate BerElement\n");
			return(FALSE);
		}

		/*
		 * Print "Sequence Start" tag to BerElement.
		 */
		Status=(ULONG)ber_printf( ber, "{" );
		DEBUGMSG("      ber_printf returned \"%S\"\n", ldap_err2string(Status));
		if(-1 == Status)
		{
			DEBUGMSG("      Error in ber_printf( ber, \"{\" )\n");
			return(FALSE);
		}

		/*
		 * Print DN tag to BerElement.
		 */
		Status=(ULONG)ber_printf( ber, "ts", LDAP_TAG_EXOP_X_MODIFY_PASSWD_ID, pcDN );
		DEBUGMSG("      ber_printf returned \"%S\"\n", ldap_err2string(Status));
		if(-1 == Status)
		{
			DEBUGMSG("      Error in ber_printf( ber, \"ts\", LDAP_TAG_EXOP_X_MODIFY_PASSWD_ID, pDN )\n");
			return(FALSE);
		}

		/*
		 * Print new user password to BerElement.
		 */
		Status=(ULONG)ber_printf( ber, "ts", LDAP_TAG_EXOP_X_MODIFY_PASSWD_NEW, pcPwd );
		DEBUGMSG("      ber_printf returned \"%S\"\n", ldap_err2string(Status));
		if(-1 == Status)
		{
			DEBUGMSG("      Error in ber_printf( ber, \"ts\", LDAP_TAG_EXOP_X_MODIFY_PASSWD_NEW, pcPwd)\n");
			return(FALSE);
		}

		/*
		 * Print "Sequence End" tag to BerElement
		 */
		Status=(ULONG)ber_printf( ber, "}" );
		DEBUGMSG("      ber_printf returned \"%S\"\n", ldap_err2string(Status));
		if(-1 == Status)
		{
			DEBUGMSG("      Error in ber_printf( ber, \"}\" )\n");
			return(FALSE);
		}

		/*
		 * Flatten the BerElement structure for xmit.
		 */
		Status = ber_flatten( ber, &bv );
		DEBUGMSG("      ber_flatten returned \"%S\"\n", ldap_err2string(Status));
		if(-1 == Status)
		{
			DEBUGMSG("      Error in ber_flatten( ber, &bv )\n");
			return(FALSE);
		}

		/*
		 * Free allocated memory.
		 */
		ber_free(ber, 1);
	
		DEBUGMSG("      bv_len        = %u\n", bv->bv_len);
		DEBUG_DUMP_MEMORY(bv->bv_val, bv->bv_len);

		/*
		* Execute the extended LDAPv3 operation.
		*/
		Status = ldap_extended_operation( ld,							// ld
										  LDAP_EXOP_X_MODIFY_PASSWD,	// OID
										  bv,							// Data
										  NULL,							// ServerControls
										  NULL,							// ClientControls
										  &id );						// MessageNumber

		DEBUGMSG("      ldap_extended_operation returned \"%S\"\n", ldap_err2string(Status));

		/*
		 * Validate the operation result.
		 */
		Status = ldap_result(ld, id, LDAP_MSG_ALL, NULL, &res);
		DEBUGMSG("      ldap_result returned \"%S\"\n", ldap_err2string(Status));
		if ( -1== Status )
		{
			DEBUGMSG("      Error in ldap_result.\n" );
			return(FALSE);
		}

		/*
		 * Terminate the extended operation.
		 */
		Status=ldap_close_extended_op(ld, id);
		DEBUGMSG("      ldap_close_extended_op returned \"%S\"\n", ldap_err2string(Status));

		/*
		 * Free allocated memory.
		 */
		if(NULL != pcDN)  LocalFree(pcDN);
		if(NULL != pcPwd) LocalFree(pcPwd);

		DEBUGMSG("   -LdapSetUserPassword\n");
	}

	return(TRUE);
}

/*************************************************************************************************
 ** Function  : LdapFreeMemory
 ** Purpose   : Fre previously allocated memory
 ** Author    : O. Dengler
 ** History   : 03/09/2002 : 1st
 ************************************************************************************************/
BOOL LdapFreeMemory(PWSTR pDN)
{
	if(NULL != pDN) ldap_memfree(pDN);

	return(TRUE);
}

/*************************************************************************************************
 ** Function  : LdapClose
 ** Purpose   : Close the server connection
 ** Author    : O. Dengler
 ** History   : 03/09/2002 : 1st
 ************************************************************************************************/
BOOL LdapClose(LDAP* ld)
{
	DEBUGMSG("   +LdapClose\n");

	/*
	 * close the connection
	 */
	ldap_unbind(ld);
	DEBUGMSG("Closed the connection to the server.\n");

	DEBUGMSG("   -LdapClose\n");
	return(TRUE);
}

/*************************************************************************************************
 ** Function  : LdapGetProtocolVersion
 ** Purpose   : Get the maximum LDAP protocol version supported by the server.
 ** Author    : O. Dengler
 ** History   : 03/20/2002 : 1st
 ************************************************************************************************/
int LdapGetProtocolVersion(void)
{
	LDAP* ld=NULL;
	LDAPMessage* pMsg=NULL;
	BerElement *ber=NULL;
	int iLdapVersion=LDAP_VERSION2;
	ULONG Status;
	PTSTR pAttrs[2] = { _T("supportedLDAPVersion" ), NULL};
	PTSTR attr=NULL;
	PTSTR *pVal=NULL;

	DEBUGMSG("   +LdapGetProtocolVersion\n");

	/*
	 * Since this code is being executed during startup, the networking
	 * subsystem could not yet be available. Try to contact the LDAP server
	 * 10 times. Wait one second between each attempt. If the server is not
	 * reachable after 10 seconds assume a serious problem and fail.
	 */
	for(int iTrials=10; (iTrials > 0) && (ld == NULL); iTrials--)
	{
		DEBUGMSG("      ldap_open(%S) (Trial #%u)\n", pLdapServer, 11-iTrials);
		ld=ldap_open(pLdapServer, LDAP_PORT);
		if(NULL == ld) Sleep(10000);
	}

	/*
	 * The LDAP server could not be contacted
	 */
	if(NULL == ld)
	{
		DEBUGMSG("   Error: unable to contact LDAP server.\n");
		return(LDAP_VERSION2);
	}

	
	/*
	 * bind to the directory
	 */
	Status=ldap_simple_bind_s(ld, pBindDN, pBindPW);
	DEBUGMSG("      ldap_simple_bind_s returned \"%S\"\n", ldap_err2string(Status));

	/*
	 * Search the root DSE for the supported LDAP version(s).
	 */
	Status=ldap_search_s(ld,
						 _T(""),
						 LDAP_SCOPE_BASE ,
						 _T("objectclass=*"),
						 pAttrs,
						 FALSE,
						 &pMsg);
	DEBUGMSG("      ldap_search_s returned \"%S\"\n", ldap_err2string(Status));

	/*
	 * Check if the server returned the "supportedLDAPVersion" attribute
	 */
	attr=ldap_first_attribute(ld, pMsg, &ber);
	if(NULL != attr)
	{
		if(!_tcscmp(attr, _T("supportedLDAPVersion")))
		{
	
			DEBUGMSG("      Server supports LDAP version ");
			pVal=ldap_get_values(ld, pMsg, attr);

			/*
			 * Enumerate the supported protocol versions and select the
			 * highest supported one.
			 */
			for(int n=0; pVal[n]; n++)
			{
				iLdapVersion=max(iLdapVersion, _ttoi(pVal[n]));
#ifdef _DEBUG
				if(NULL != pVal[n+1])
					DEBUGMSG("%S, ", pVal[n]);
				else
					DEBUGMSG("%S\n", pVal[n]);
#endif
			}
	
		}
		else
		{
			DEBUGMSG("      Error: illegal attribute name %S\n", attr);
		}
	}
	else
	{
		DEBUGMSG("      Error: NULL attribute\n");
	}


	DEBUGMSG("      Selected LDAP Version %u\n", iLdapVersion);

	/*
	 * Bye
	 */
	LdapClose(ld);

	DEBUGMSG("   -LdapGetProtocolVersion\n");

	return(iLdapVersion);
}